"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const mql_worker_1 = require("@mavenomics/mql-worker");
const coreutils_1 = require("@mavenomics/coreutils");
const coreutils_2 = require("@phosphor/coreutils");
const CodeMirror = require("codemirror");
const bindings_1 = require("@mavenomics/bindings");
require("codemirror/addon/lint/lint");
require("codemirror/addon/hint/show-hint");
const worker = new mql_worker_1.WorkerWrapper();
const workerMsgs = {};
worker.onMessage.subscribe((msgData) => {
    const { id } = msgData;
    if (msgData.type === "parseQueryResult" && id) {
        const data = msgData.data;
        const msg = workerMsgs[id];
        if (!msg) {
            console.error("Received bad parse result", msg, id, data);
            return;
        }
        try {
            msg.resolve(JSON.parse(data));
        }
        catch (e) {
            msg.reject(e);
        }
        finally {
            delete workerMsgs[id];
        }
    }
    else if (msgData.type === "getAvailableFunctionsResult" && id) {
        const data = msgData.data;
        const msg = workerMsgs[id];
        if (!msg) {
            console.error("Received bad parse result", msg, id, data);
            return;
        }
        try {
            const funcsJson = JSON.parse(data);
            const infos = funcsJson.map(i => {
                return {
                    ...i,
                    returnType: coreutils_1.Types.findType(i.returnTypeName) || coreutils_1.Types.Any,
                    args: i.args.map(arg => {
                        return {
                            ...arg,
                            type: coreutils_1.Types.findType(arg.typeName) || coreutils_1.Types.Any
                        };
                    })
                };
            });
            msg.resolve(infos);
        }
        catch (e) {
            msg.reject(e);
        }
        finally {
            delete workerMsgs[id];
        }
    }
});
function parseQuery(query) {
    const id = coreutils_2.UUID.uuid4();
    let resolve;
    let reject;
    let promise = new Promise((res, rej) => { resolve = res; reject = rej; });
    workerMsgs[id] = { id: id, reject: reject, resolve: resolve };
    worker.postMessage({
        type: "parseQuery" /* ParseQueryRequest */,
        id: id,
        codeText: query
    });
    return promise;
}
function getAvailableFunctions() {
    const id = coreutils_2.UUID.uuid4();
    let resolve;
    let reject;
    let promise = new Promise((res, rej) => { resolve = res; reject = rej; });
    workerMsgs[id] = { id: id, reject: reject, resolve: resolve };
    worker.postMessage({
        type: "getAvailableFunctions" /* GetFunctionsRequest */,
        id: id
    });
    return promise;
}
exports.getAvailableFunctions = getAvailableFunctions;
const mqlLinter = (text, resolve, config, cm) => {
    parseQuery(text).then(data => {
        //Expose the sets/defs to the autocomplete
        cm.state.MqlSetsDefs = data;
        const errors = data.errors;
        if (errors.length > 0) {
            //We only care about the first error since the other errors are typically from recovery mode.
            const error = errors[0];
            let tokenLength = error.token && error.token.text ? error.token.text.length : 0;
            resolve([{
                    message: error.msg,
                    severity: "error",
                    from: CodeMirror.Pos(error.line - 1, error.column),
                    to: CodeMirror.Pos(error.line - 1, error.column + tokenLength)
                }]);
        }
        else {
            resolve([]);
        }
    }).catch((err) => {
        console.error("Unhandled parse error", err);
        resolve([]);
    });
};
mqlLinter.async = true; //In this mode the linter will throw away stale lint results.
CodeMirror.registerHelper("lint", "mql", mqlLinter);
function mqlTrim(str) {
    //Remove whitespace to handle the token under the cursor being only whitespace.
    //Also remove leading '.'. This is needed since the codemirror mql tokenizer isn't complete.
    str = str.replace(/^[\.\s]*/g, "");
    return str;
}
const getReferencedGlobals = (text) => {
    const comment = bindings_1.getCommentLine(text);
    return comment != null ? bindings_1.getGlobalsFromComment(comment) : null;
};
const generateMqlSignature = (func) => {
    //Mql functions always start with a row arg.
    const args = func.args.slice(1);
    //Todo: Decide on a signature format.
    //Currently mql doesn't support defaults(optional args).
    //When that changes we will want to pick a better signature format.
    const requiredArgs = args;
    const repeatingArgs = args.slice(args.length - func.repeatingArgs);
    const requiredText = requiredArgs.map((a) => a.name + ": " + a.type.name).join(", ");
    const leadingComma = requiredArgs.length > 0 ? ", " : "";
    const repeatingText = repeatingArgs.length > 0 ?
        ` [${leadingComma}${repeatingArgs.map((a) => a.name + ": " + a.type.name).join(", ")}]+` :
        "";
    const signature = `(${requiredText}${repeatingText}): ${func.returnTypeName}`;
    return signature;
};
const generateDefSignature = (def) => {
    return `(${def.paramNames.map((e) => `@${e}`).join(", ")})`;
};
const mqlRenderHint = (element, self, data) => {
    //self is the object containing {list, from, to, render, etc}.
    //data is the current completion from {list} being rendered.
    const containerEl = document.createElement("div");
    containerEl.className = "CodeMirror-hint-container";
    const nameEl = document.createElement("div");
    nameEl.className = "CodeMirror-hint-name";
    const sigEl = document.createElement("div");
    sigEl.className = "CodeMirror-hint-signature";
    nameEl.appendChild(document.createTextNode(data.name));
    sigEl.appendChild(document.createTextNode(data.signature || "")); //signature is optional
    containerEl.appendChild(nameEl);
    containerEl.appendChild(sigEl);
    element.appendChild(containerEl);
};
const mqlHints = async (cm, resolve, config) => {
    //Todo: cache getAvailableFunctions since mqlHints is called as you type with the autocomplete dialog open
    const availableFuncs = await getAvailableFunctions();
    const options = cm.options;
    const globalsService = options && options.context ? options.context.globals : null;
    const allGlobals = globalsService ? Array.from(globalsService).map((g) => g.name) : [];
    const globals = getReferencedGlobals(cm.getDoc().getValue()) || allGlobals;
    const cur = cm.getDoc().getCursor();
    let token = cm.getTokenAt(cur);
    //Remove any of the token that appears after the cursor.
    //This allows you to autocomplete mid word.
    if (token.end > cur.ch) {
        token.end = cur.ch;
        token.string = token.string.slice(0, cur.ch - token.start);
    }
    let setsDefs = cm.state.MqlSetsDefs || { sets: [], defs: [] };
    //Combine the functions, sets, defs and reference globals into a MqlCompletion array
    let mqlFuncs = availableFuncs.map(fi => ({ name: fi.name, signature: generateMqlSignature(fi) }));
    let sets = setsDefs.sets.map((e) => ({ name: `@${e}` }));
    //TODO: Display def signatures
    let defs = setsDefs.defs.map((e) => ({ name: `@${e.name}`, signature: generateDefSignature(e) }));
    let globalAtNames = globals.map((e) => ({ name: `@${e}` }));
    let completions = mqlFuncs.concat(sets, defs, globalAtNames);
    //Find functions that match the token.
    let word = mqlTrim(token.string);
    if (word.length > 0) {
        completions = completions.filter(fi => fi.name.toLowerCase().startsWith(word.toLowerCase()));
    }
    completions.sort((a, b) => a.name.localeCompare(b.name));
    let replaceToken = word.length > 0;
    resolve({
        list: completions.map(fi => ({ name: fi.name, signature: fi.signature, render: mqlRenderHint, text: fi.name })),
        from: replaceToken ? CodeMirror.Pos(cur.line, cur.ch - word.length) : cur,
        to: cur,
    });
};
mqlHints.async = true;
CodeMirror.registerHelper("hint", "mql", mqlHints);
